import base64
import hashlib
import hmac
import json
# import ssl
# import threading
from datetime import datetime
from time import mktime
from urllib.parse import urlencode, urlparse
from wsgiref.handlers import format_date_time

import websocket
from websocket import create_connection

import _thread as thread

from .types import ChatInput, ChatOutput, Choice, Message, Usage
from .utils import get_ts


class Ws_Param(object):
    # 初始化
    def __init__(self, APPID, APIKey, APISecret, gpt_url):
        self.APPID = APPID
        self.APIKey = APIKey
        self.APISecret = APISecret
        self.host = urlparse(gpt_url).netloc
        self.path = urlparse(gpt_url).path
        self.gpt_url = gpt_url

    # 生成url
    def create_url(self):
        # 生成RFC1123格式的时间戳
        now = datetime.now()
        date = format_date_time(mktime(now.timetuple()))

        # 拼接字符串
        signature_origin = 'host: ' + self.host + '\n'
        signature_origin += 'date: ' + date + '\n'
        signature_origin += 'GET ' + self.path + ' HTTP/1.1'

        # 进行hmac-sha256进行加密
        signature_sha = hmac.new(
            self.APISecret.encode('utf-8'),
            signature_origin.encode('utf-8'),
            digestmod=hashlib.sha256).digest()

        signature_sha_base64 = base64.b64encode(
            signature_sha).decode(encoding='utf-8')

        authorization_origin = (
          f'api_key="{self.APIKey}", '
          f'algorithm="hmac-sha256", headers="host date request-line",'
          f' signature="{signature_sha_base64}"')

        authorization = base64.b64encode(
            authorization_origin.encode('utf-8')).decode(encoding='utf-8')

        # 将请求的鉴权参数组合为字典
        v = {
            'authorization': authorization,
            'date': date,
            'host': self.host
        }
        # 拼接鉴权参数，生成url
        url = self.gpt_url + '?' + urlencode(v)
        # 此处打印出建立连接时候的url,参考本demo的时候可取消上方打印的注释，
        # 比对相同参数时生成的url与自己代码生成的url是否一致
        return url


# 收到websocket错误的处理
def on_error(ws, error):
    print('### error:', error)


# 收到websocket关闭的处理
def on_close(ws):
    print('### closed ###')


# 收到websocket连接建立的处理
def on_open(ws):
    thread.start_new_thread(run, (ws,))


def run(ws, *args):
    data = json.dumps(gen_params(appid=ws.appid, question=ws.question))
    ws.send(data)


# 收到websocket消息的处理
def on_message(ws, message):
    print(message)
    data = json.loads(message)
    code = data['header']['code']
    if code != 0:
        print(f'请求错误: {code}, {data}')
        ws.close()
    else:
        choices = data['payload']['choices']
        status = choices['status']
        content = choices['text'][0]['content']
        print(content, end='')
        if status == 2:
            ws.close()


def gen_params(appid, question):
    data = {
        'header': {
            'app_id': appid,
            'uid': '1234'
        },
        'parameter': {
            'chat': {
                'domain': 'general',
                'random_threshold': 0.5,
                'max_tokens': 2048,
                'auditing': 'default'
            }
        },
        'payload': {
            'message': {
                'text': [
                    {'role': 'user', 'content': question}
                ]
            }
        }
    }
    return data


class ChatCompletion(object):
    def __init__(self, appid, api_key, api_secret, **kwargs):
        gpt_url1 = 'ws://spark-api.xf-yun.com/v1.1/chat'
        gpt_url2 = 'ws://spark-api.xf-yun.com/v2.1/chat'
        gpt_url3 = 'ws://spark-api.xf-yun.com/v3.1/chat'
        self.wsParam1 = Ws_Param(appid, api_key, api_secret, gpt_url1)
        self.wsParam2 = Ws_Param(appid, api_key, api_secret, gpt_url2)
        self.wsParam3 = Ws_Param(appid, api_key, api_secret, gpt_url3)

        websocket.enableTrace(False)

        # todo: modify to the ws pool
        # self.ws = websocket.WebSocket()
        # self.ws.connect(self.wsUrl)

        # self.mutex = threading.Lock()
        self.header = {'app_id': appid, 'uid': 'elem'}

    def __del__(self):
        pass
        # self.ws.close()

    def __call__(self, inp: ChatInput, verbose=False):
        messages = inp.messages
        model = inp.model
        # top_p = 0.7 if inp.top_p is None else inp.top_p
        temperature = 0.5 if inp.temperature is None else inp.temperature
        # stream = False if inp.stream is None else inp.stream
        max_tokens = 1024 if inp.max_tokens is None else inp.max_tokens
        # stop = None
        # if inp.stop is not None:
        #     stop = inp.stop.split('||')

        new_messages = []
        for m in messages:
            role = m.role
            if role == 'system':
                role = 'user'
            new_messages.append({'role': role, 'content': m.content})

        domain = 'general'
        if model == 'spark-v2.0':
            domain = 'generalv2'
        elif model == 'spark-v3.0':
            domain = 'generalv3'

        created = get_ts()
        payload = {
            'header': self.header,
            'payload': {'message': {'text': new_messages}},
            'parameter': {
              'chat': {
                'domain': domain,
                'temperature': temperature,
                'max_tokens': max_tokens,
                'auditing': 'default'
              }
            }
        }

        if verbose:
            print('payload', payload)

        req_type = 'chat.completion'
        status_code = 200
        status_message = 'success'
        choices = []
        usage = Usage()
        texts = []
        ws = None
        try:
            # self.mutex.acquire()
            if model == 'spark-v2.0':
                wsUrl = self.wsParam2.create_url()
            elif model == 'spark-v3.0':
                wsUrl = self.wsParam3.create_url()
            else:
                wsUrl = self.wsParam1.create_url()

            ws = create_connection(wsUrl)
            ws.send(json.dumps(payload))
            texts = []
            while True:
                raw_data = ws.recv()
                if not raw_data:
                    break
                resp = json.loads(raw_data)
                if resp['header']['code'] == 0:
                    texts.append(
                        resp['payload']['choices']['text'][0]['content'])
                if resp['header']['code'] == 0 and resp['header']['status'] == 2:
                    usage_dict = resp['payload']['usage']['text']
                    usage_dict.pop('question_tokens')
                    usage = Usage(**usage_dict)
        except Exception as e:
            print('exception', e)
            status_code = 401
            status_message = str(e)
        finally:
            if ws:
                ws.close()
            # self.mutex.release()

        if texts:
            finish_reason = 'default'
            msg = Message(role='assistant', content=''.join(texts))
            cho = Choice(index=0, message=msg,
                         finish_reason=finish_reason)
            choices.append(cho)

        return ChatOutput(
            status_code=status_code,
            status_message=status_message,
            model=model, object=req_type, created=created,
            choices=choices, usage=usage)



SYS_MSG = """
你是一个善于分析用户需求的专家，你非常擅长根据用户对任务的描述，从你的工具箱中选择合适的工具给用户。
你的工具箱中有以下工具，其中tool_name表示这个工具的名字，tool_description表示工具的用途：{tool_pool}。

在选择工具时，你需要遵守以下规则：
1. 你需要根据用户输入的任务描述，选择最符合用户需求的工具，不要随意选择工具给用户。
2. 只能从你的工具箱中选择工具，一定不要自己编造工具箱中不存在的工具给用户！
3. 你只需要返回工具的名称，不需要返回工具的描述以及其他额外的信息，返回的格式为：['tool1', 'tool2', 'tool3']。
4. 如果你觉得没有合适的工具，请返回：[]。
5. 当任务描述中出现“Skill”时，你可以重点关注“Skill”中的内容，这些内容可能会帮助你更好地选择工具。
6. 如果你回答对了，将会获得100块钱的奖励，如果回答错误，将会扣除50块钱。

以下是一些示例:
例子1:
任务名: 论文助手
任务描述: 帮助学生查找相关论文
建议的工具: ['arxiv', 'bing_search']

例子2:
任务名: 时间查询
任务描述: 查询当前时间
建议的工具: ['get_current_time']
"""

HUMAN_MSG = """
任务名: {task_name}
任务描述: {task_description}
建议的工具: 
"""



import math
from math import *

import sympy
from langchain.pydantic_v1 import BaseModel, Field
from langchain.tools import tool
from sympy import *


class CalculatorInput(BaseModel):
    expression: str = Field(
        description="The input to this tool should be a mathematical expression using only Python's built-in mathematical operators.",
        example='200*7',
    )


@tool("calculator", args_schema=CalculatorInput)
def calculator(expression):
    """Useful to perform any mathematical calculations,
    like sum, minus, multiplication, division, etc
    """
    try:
        return eval(expression)
    except SyntaxError:
        return "Error: Invalid syntax in mathematical expression"



"""Chain for interacting with SQL Database."""
from __future__ import annotations

from typing import Any, Dict, List, Optional

from langchain.callbacks.manager import AsyncCallbackManagerForChainRun, CallbackManagerForChainRun
from langchain.chains.llm import LLMChain
from langchain.prompts.prompt import PromptTemplate
from langchain.tools.sql_database.prompt import QUERY_CHECKER
from langchain_experimental.sql import SQLDatabaseChain as SQLDatabaseChainExperimental

INTERMEDIATE_STEPS_KEY = 'intermediate_steps'


class SQLDatabaseChain(SQLDatabaseChainExperimental):
    """Chain for interacting with SQL Database.

    Example:
        .. code-block:: python

            from langchain_experimental.sql import SQLDatabaseChain
            from langchain.llms import OpenAI, SQLDatabase
            db = SQLDatabase(...)
            db_chain = SQLDatabaseChain.from_llm(OpenAI(), db)

    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include the permissions this chain needs.
        Failure to do so may result in data corruption or loss, since this chain may
        attempt commands like `DROP TABLE` or `INSERT` if appropriately prompted.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this chain.
        This issue shows an example negative outcome if these steps are not taken:
        https://github.com/langchain-ai/langchain/issues/5923
    """

    async def _acall(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        input_text = f'{inputs[self.input_key]}\nSQLQuery:'
        await _run_manager.on_text(input_text, verbose=self.verbose)
        # If not present, then defaults to None which is all tables.
        table_names_to_use = inputs.get('table_names_to_use')
        table_info = self.database.get_table_info(table_names=table_names_to_use)
        llm_inputs = {
            'input': input_text,
            'top_k': str(self.top_k),
            'dialect': self.database.dialect,
            'table_info': table_info,
            'stop': ['\nSQLResult:'],
        }
        if self.memory is not None:
            for k in self.memory.memory_variables:
                llm_inputs[k] = inputs[k]
        intermediate_steps: List = []
        try:
            intermediate_steps.append(llm_inputs.copy())  # input: sql generation
            sql_cmd = await self.llm_chain.apredict(
                callbacks=_run_manager.get_child(),
                **llm_inputs,
            )
            sql_cmd = sql_cmd.strip()
            if self.return_sql:
                return {self.output_key: sql_cmd}
            if not self.use_query_checker:
                await _run_manager.on_text(sql_cmd, color='green', verbose=self.verbose)
                intermediate_steps.append(sql_cmd)  # output: sql generation (no checker)
                intermediate_steps.append({'sql_cmd': sql_cmd})  # input: sql exec
                result = self.database.run(sql_cmd)
                intermediate_steps.append(str(result))  # output: sql exec
            else:
                query_checker_prompt = self.query_checker_prompt or PromptTemplate(
                    template=QUERY_CHECKER, input_variables=['query', 'dialect'])
                query_checker_chain = LLMChain(llm=self.llm_chain.llm, prompt=query_checker_prompt)
                query_checker_inputs = {
                    'query': sql_cmd,
                    'dialect': self.database.dialect,
                }
                checked_sql_command: str = await query_checker_chain.apredict(
                    callbacks=_run_manager.get_child(), **query_checker_inputs)
                checked_sql_command = checked_sql_command.strip()
                intermediate_steps.append(checked_sql_command)  # output: sql generation (checker)
                await _run_manager.on_text(checked_sql_command,
                                           color='green',
                                           verbose=self.verbose)
                intermediate_steps.append({'sql_cmd': checked_sql_command})  # input: sql exec
                result = self.database.run(checked_sql_command)
                intermediate_steps.append(str(result))  # output: sql exec
                sql_cmd = checked_sql_command

            await _run_manager.on_text('\nSQLResult: ', verbose=self.verbose)
            await _run_manager.on_text(result, color='yellow', verbose=self.verbose)
            # If return direct, we just set the final result equal to
            # the result of the sql query result, otherwise try to get a human readable
            # final answer
            if self.return_direct:
                final_result = result
            else:
                await _run_manager.on_text('\nAnswer:', verbose=self.verbose)
                input_text += f'{sql_cmd}\nSQLResult: {result}\nAnswer:'
                llm_inputs['input'] = input_text
                intermediate_steps.append(llm_inputs.copy())  # input: final answer
                final_result = await self.llm_chain.apredict(
                    callbacks=_run_manager.get_child(),
                    **llm_inputs,
                )
                final_result = final_result.strip()
                intermediate_steps.append(final_result)  # output: final answer
                await _run_manager.on_text(final_result, color='green', verbose=self.verbose)
            chain_result: Dict[str, Any] = {self.output_key: final_result}
            if self.return_intermediate_steps:
                chain_result[INTERMEDIATE_STEPS_KEY] = intermediate_steps
            return chain_result
        except Exception as exc:
            # Append intermediate steps to exception, to aid in logging and later
            # improvement of few shot prompt seeds
            exc.intermediate_steps = intermediate_steps  # type: ignore
            raise exc



import json
import time
from typing import Any

import pandas as pd
import requests
from langchain.pydantic_v1 import BaseModel, Field
from langchain_core.tools import BaseTool

from .base import MultArgsSchemaTool


class QueryArg(BaseModel):
    start_date: str = Field(default='', description='开始月份, 使用YYYY-MM-DD 方式表示', example='2023-01-01')
    end_date: str = Field(default='', description='结束月份，使用YYYY-MM-DD 方式表示', example='2023-05-01')


class MacroData(BaseModel):

    @classmethod
    def china_shrzgm(cls, start_date: str = '', end_date: str = '') -> pd.DataFrame:
        """中国社会融资规模增量月度统计数据。\
返回月份，社会融资规模增量(单位：亿元)，\
以及社融分项包括：人民币贷款，外币贷款，委托贷款，\
信托贷款，未贴现银行承兑汇票，企业债券，非金融企业境内股票融资
        """
        url = 'http://data.mofcom.gov.cn/datamofcom/front/gnmy/shrzgmQuery'
        r = requests.post(url)
        data_json = r.json()
        temp_df = pd.DataFrame(data_json)
        temp_df.columns = [
            '月份',
            '其中-未贴现银行承兑汇票',
            '其中-委托贷款',
            '其中-委托贷款外币贷款',
            '其中-人民币贷款',
            '其中-企业债券',
            '社会融资规模增量',
            '其中-非金融企业境内股票融资',
            '其中-信托贷款',
        ]
        temp_df = temp_df[
            [
                '月份',
                '社会融资规模增量',
                '其中-人民币贷款',
                '其中-委托贷款外币贷款',
                '其中-委托贷款',
                '其中-信托贷款',
                '其中-未贴现银行承兑汇票',
                '其中-企业债券',
                '其中-非金融企业境内股票融资',
            ]
        ]
        temp_df['社会融资规模增量'] = pd.to_numeric(temp_df['社会融资规模增量'], errors='coerce')
        temp_df['其中-人民币贷款'] = pd.to_numeric(temp_df['其中-人民币贷款'], errors='coerce')
        temp_df['其中-委托贷款外币贷款'] = pd.to_numeric(temp_df['其中-委托贷款外币贷款'], errors='coerce')
        temp_df['其中-委托贷款'] = pd.to_numeric(temp_df['其中-委托贷款'], errors='coerce')
        temp_df['其中-信托贷款'] = pd.to_numeric(temp_df['其中-信托贷款'], errors='coerce')
        temp_df['其中-未贴现银行承兑汇票'] = pd.to_numeric(temp_df['其中-未贴现银行承兑汇票'], errors='coerce')
        temp_df['其中-企业债券'] = pd.to_numeric(temp_df['其中-企业债券'], errors='coerce')
        temp_df['其中-非金融企业境内股票融资'] = pd.to_numeric(temp_df['其中-非金融企业境内股票融资'], errors='coerce')
        temp_df.sort_values(['月份'], inplace=True)
        if start_date and end_date:
            start = start_date.split('-')[0] + start_date.split('-')[1]
            end = end_date.split('-')[0] + end_date.split('-')[1]
            temp_df = temp_df[(temp_df['月份'] >= start) & (temp_df['月份'] <= end)]

        temp_df.reset_index(drop=True, inplace=True)
        return temp_df.to_markdown()

    # 金十数据中心-经济指标-中国-国民经济运行状况-经济状况-中国GDP年率报告
    @classmethod
    def china_gdp_yearly(cls, start_date: str = '', end_date: str = '') -> pd.DataFrame:
        """中国国内生产总值（GDP）季度统计数据。\
返回当年累计季度，GDP 绝对值（单位：亿元），同比增长（单位：%），\
第一产业 GDP 绝对值，（单位：亿元），第一产业同比增长（单位：%），\
第二产业 GDP 绝对值，（单位：亿元），第二产业同比增长（单位：%），\
第三产业 GDP 绝对值，（单位：亿元），第三产业同比增长（单位：%）
        """
        JS_CHINA_GDP_YEARLY_URL = 'https://cdn.jin10.com/dc/reports/dc_chinese_gdp_yoy_all.js?v={}&_={}'
        t = time.time()
        r = requests.get(JS_CHINA_GDP_YEARLY_URL.format(str(int(round(t * 1000))), str(int(round(t * 1000)) + 90)))
        json_data = json.loads(r.text[r.text.find('{'): r.text.rfind('}') + 1])
        date_list = [item['date'] for item in json_data['list']]
        value_list = [item['datas']['中国GDP年率报告'] for item in json_data['list']]
        value_df = pd.DataFrame(value_list)
        value_df.columns = json_data['kinds']
        value_df.index = pd.to_datetime(date_list)
        temp_df = value_df['今值(%)']
        url = 'https://datacenter-api.jin10.com/reports/list_v2'
        params = {
            'max_date': '',
            'category': 'ec',
            'attr_id': '57',
            '_': str(int(round(t * 1000))),
        }
        headers = {
            'accept': '*/*',
            'accept-encoding': 'gzip, deflate, br',
            'accept-language': 'zh-CN,zh;q=0.9,en;q=0.8',
            'cache-control': 'no-cache',
            'origin': 'https://datacenter.jin10.com',
            'pragma': 'no-cache',
            'referer': 'https://datacenter.jin10.com/reportType/dc_usa_michigan_consumer_sentiment',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-site',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36',  # noqa
            'x-app-id': 'rU6QIu7JHe2gOUeR',
            'x-csrf-token': '',
            'x-version': '1.0.0',
        }
        r = requests.get(url, params=params, headers=headers)
        temp_se = pd.DataFrame(r.json()['data']['values']).iloc[:, :2]
        temp_se.index = pd.to_datetime(temp_se.iloc[:, 0])
        temp_se = temp_se.iloc[:, 1]
        temp_df = pd.concat([temp_df, temp_se])
        temp_df.dropna(inplace=True)
        temp_df.sort_index(inplace=True)
        temp_df = temp_df.reset_index()
        temp_df.columns = ['date', 'value']
        # temp_df['date'] = pd.to_datetime(temp_df['date']).dt.date
        temp_df['value'] = pd.to_numeric(temp_df['value'])
        temp_df = temp_df.drop_duplicates()
        if start_date and end_date:
            temp_df = temp_df[(temp_df['date'] >= start_date) & (temp_df['date'] <= end_date)]
        return temp_df.to_markdown()

    @classmethod
    def china_cpi(cls, start_date: str = '', end_date: str = '') -> pd.DataFrame:
        """中国居民消费价格指数(CPI，上年同月=100)月度统计数据。\
返回月份，全国当月 CPI，全国当月同比增长，全国当月环比增长，全国当年 CPI 累计值；\
城市当月 CPI，城市当月同比增长，城市当月环比增长，城市当年 CPI 累计值；\
农村当月 CPI，农村当月同比增长，农村当月环比增长，农村当年 CPI 累计值。
        """
        url = 'https://datacenter-web.eastmoney.com/api/data/v1/get'
        params = {
            'columns': 'REPORT_DATE,TIME,NATIONAL_SAME,NATIONAL_BASE,NATIONAL_SEQUENTIAL,NATIONAL_ACCUMULATE,CITY_SAME,CITY_BASE,CITY_SEQUENTIAL,CITY_ACCUMULATE,RURAL_SAME,RURAL_BASE,RURAL_SEQUENTIAL,RURAL_ACCUMULATE',  # noqa
            'pageNumber': '1',
            'pageSize': '2000',
            'sortColumns': 'REPORT_DATE',
            'sortTypes': '-1',
            'source': 'WEB',
            'client': 'WEB',
            'reportName': 'RPT_ECONOMY_CPI',
            'p': '1',
            'pageNo': '1',
            'pageNum': '1',
            '_': '1669047266881',
        }
        r = requests.get(url, params=params)
        data_json = r.json()
        temp_df = pd.DataFrame(data_json['result']['data'])
        temp_df.columns = [
            '-',
            '月份',
            '全国-同比增长',
            '全国-当月',
            '全国-环比增长',
            '全国-累计',
            '城市-同比增长',
            '城市-当月',
            '城市-环比增长',
            '城市-累计',
            '农村-同比增长',
            '农村-当月',
            '农村-环比增长',
            '农村-累计',
        ]
        temp_df = temp_df[
            [
                '月份',
                '全国-当月',
                '全国-同比增长',
                '全国-环比增长',
                '全国-累计',
                '城市-当月',
                '城市-同比增长',
                '城市-环比增长',
                '城市-累计',
                '农村-当月',
                '农村-同比增长',
                '农村-环比增长',
                '农村-累计',
            ]
        ]
        temp_df['全国-当月'] = pd.to_numeric(temp_df['全国-当月'], errors='coerce')
        temp_df['全国-同比增长'] = pd.to_numeric(temp_df['全国-同比增长'], errors='coerce')
        temp_df['全国-环比增长'] = pd.to_numeric(temp_df['全国-环比增长'], errors='coerce')
        temp_df['全国-累计'] = pd.to_numeric(temp_df['全国-累计'], errors='coerce')
        temp_df['城市-当月'] = pd.to_numeric(temp_df['城市-当月'], errors='coerce')
        temp_df['城市-同比增长'] = pd.to_numeric(temp_df['城市-同比增长'], errors='coerce')
        temp_df['城市-环比增长'] = pd.to_numeric(temp_df['城市-环比增长'], errors='coerce')
        temp_df['城市-累计'] = pd.to_numeric(temp_df['城市-累计'], errors='coerce')
        temp_df['农村-当月'] = pd.to_numeric(temp_df['农村-当月'], errors='coerce')
        temp_df['农村-同比增长'] = pd.to_numeric(temp_df['农村-同比增长'], errors='coerce')
        temp_df['农村-环比增长'] = pd.to_numeric(temp_df['农村-环比增长'], errors='coerce')
        temp_df['农村-累计'] = pd.to_numeric(temp_df['农村-累计'], errors='coerce')
        if start_date and end_date:
            start = start_date.split('-')[0] + '年' + start_date.split('-')[1] + '月份'
            end = end_date.split('-')[0] + '年' + end_date.split('-')[1] + '月份'
            temp_df = temp_df[(temp_df['月份'] >= start) & (temp_df['月份'] <= end)]

        return temp_df.to_markdown()

    @classmethod
    def china_ppi(cls, start_date: str = '', end_date: str = '') -> pd.DataFrame:
        """中国工业品出厂价格指数（PPI）月度统计数据。返回月份，当月 PPI，当月同比增长，当年 CPI 累计值。"""
        url = 'https://datacenter-web.eastmoney.com/api/data/v1/get'
        params = {
            'columns': 'REPORT_DATE,TIME,BASE,BASE_SAME,BASE_ACCUMULATE',
            'pageNumber': '1',
            'pageSize': '2000',
            'sortColumns': 'REPORT_DATE',
            'sortTypes': '-1',
            'source': 'WEB',
            'client': 'WEB',
            'reportName': 'RPT_ECONOMY_PPI',
            'p': '1',
            'pageNo': '1',
            'pageNum': '1',
            '_': '1669047266881',
        }
        r = requests.get(url, params=params)
        data_json = r.json()
        temp_df = pd.DataFrame(data_json['result']['data'])
        temp_df.columns = [
            '-',
            '月份',
            '当月',
            '当月同比增长',
            '累计',
        ]
        temp_df = temp_df[
            [
                '月份',
                '当月',
                '当月同比增长',
                '累计',
            ]
        ]
        temp_df['当月'] = pd.to_numeric(temp_df['当月'], errors='coerce')
        temp_df['当月同比增长'] = pd.to_numeric(temp_df['当月同比增长'], errors='coerce')
        temp_df['累计'] = pd.to_numeric(temp_df['累计'], errors='coerce')
        if start_date and end_date:
            start = start_date.split('-')[0] + '年' + start_date.split('-')[1] + '月份'
            end = end_date.split('-')[0] + '年' + end_date.split('-')[1] + '月份'
            temp_df = temp_df[(temp_df['月份'] >= start) & (temp_df['月份'] <= end)]
        return temp_df.to_markdown()

    @classmethod
    def china_pmi(cls, start_date: str = '', end_date: str = '') -> str:
        """中国 PMI （采购经理人指数）月度统计数据。
        返回数据包括：月份制造业 PMI，制造业 PMI 同比增长，非制造业 PMI，非制造业 PMI 同比增长。
        """
        url = "https://datacenter-web.eastmoney.com/api/data/v1/get"
        params = {
            "columns": "REPORT_DATE,TIME,MAKE_INDEX,MAKE_SAME,NMAKE_INDEX,NMAKE_SAME",
            "pageNumber": "1",
            "pageSize": "2000",
            "sortColumns": "REPORT_DATE",
            "sortTypes": "-1",
            "source": "WEB",
            "client": "WEB",
            "reportName": "RPT_ECONOMY_PMI",
            "p": "1",
            "pageNo": "1",
            "pageNum": "1",
            "_": "1669047266881",
        }
        r = requests.get(url, params=params)
        data_json = r.json()
        temp_df = pd.DataFrame(data_json["result"]["data"])
        temp_df.columns = [
            "-",
            "月份",
            "制造业-指数",
            "制造业-同比增长",
            "非制造业-指数",
            "非制造业-同比增长",
        ]
        temp_df = temp_df[
            [
                "月份",
                "制造业-指数",
                "制造业-同比增长",
                "非制造业-指数",
                "非制造业-同比增长",
            ]
        ]
        temp_df["制造业-指数"] = pd.to_numeric(temp_df["制造业-指数"], errors="coerce")
        temp_df["制造业-同比增长"] = pd.to_numeric(
            temp_df["制造业-同比增长"], errors="coerce"
        )
        temp_df["非制造业-指数"] = pd.to_numeric(temp_df["非制造业-指数"], errors="coerce")
        temp_df["非制造业-同比增长"] = pd.to_numeric(
            temp_df["非制造业-同比增长"], errors="coerce"
        )
        if start_date and end_date:
            start = start_date.split('-')[0] + '年' + start_date.split('-')[1] + '月份'
            end = end_date.split('-')[0] + '年' + end_date.split('-')[1] + '月份'
            temp_df = temp_df[(temp_df['月份'] >= start) & (temp_df['月份'] <= end)]
        return temp_df.to_markdown()

    @classmethod
    def china_money_supply(cls, start_date: str = '', end_date: str = '') -> pd.DataFrame:
        """中国货币供应量（M2，M1，M0）月度统计数据。\
返回月份，M2 数量（单位：亿元），M2 同比（单位：%），\
M2 环比（单位：%）， M1 数量（单位：亿元），\
M1 同比（单位：%），M1 环比（单位：%）， \
M0数量（单位：亿元），M0 同比（单位：%），M0 环比（单位：%）
        """
        url = 'https://datacenter-web.eastmoney.com/api/data/v1/get'
        params = {
            'columns': 'REPORT_DATE,TIME,BASIC_CURRENCY,BASIC_CURRENCY_SAME,BASIC_CURRENCY_SEQUENTIAL,CURRENCY,CURRENCY_SAME,CURRENCY_SEQUENTIAL,FREE_CASH,FREE_CASH_SAME,FREE_CASH_SEQUENTIAL',  # noqa
            'pageNumber': '1',
            'pageSize': '2000',
            'sortColumns': 'REPORT_DATE',
            'sortTypes': '-1',
            'source': 'WEB',
            'client': 'WEB',
            'reportName': 'RPT_ECONOMY_CURRENCY_SUPPLY',
            'p': '1',
            'pageNo': '1',
            'pageNum': '1',
            '_': '1669047266881',
        }
        r = requests.get(url, params=params)
        data_json = r.json()
        temp_df = pd.DataFrame(data_json['result']['data'])
        temp_df.columns = [
            '-',
            '月份',
            '货币和准货币(M2)-数量(亿元)',
            '货币和准货币(M2)-同比增长',
            '货币和准货币(M2)-环比增长',
            '货币(M1)-数量(亿元)',
            '货币(M1)-同比增长',
            '货币(M1)-环比增长',
            '流通中的现金(M0)-数量(亿元)',
            '流通中的现金(M0)-同比增长',
            '流通中的现金(M0)-环比增长',
        ]
        temp_df = temp_df[
            [
                '月份',
                '货币和准货币(M2)-数量(亿元)',
                '货币和准货币(M2)-同比增长',
                '货币和准货币(M2)-环比增长',
                '货币(M1)-数量(亿元)',
                '货币(M1)-同比增长',
                '货币(M1)-环比增长',
                '流通中的现金(M0)-数量(亿元)',
                '流通中的现金(M0)-同比增长',
                '流通中的现金(M0)-环比增长',
            ]
        ]

        temp_df['货币和准货币(M2)-数量(亿元)'] = pd.to_numeric(temp_df['货币和准货币(M2)-数量(亿元)'])
        temp_df['货币和准货币(M2)-同比增长'] = pd.to_numeric(temp_df['货币和准货币(M2)-同比增长'])
        temp_df['货币和准货币(M2)-环比增长'] = pd.to_numeric(temp_df['货币和准货币(M2)-环比增长'])
        temp_df['货币(M1)-数量(亿元)'] = pd.to_numeric(temp_df['货币(M1)-数量(亿元)'])
        temp_df['货币(M1)-同比增长'] = pd.to_numeric(temp_df['货币(M1)-同比增长'])
        temp_df['货币(M1)-环比增长'] = pd.to_numeric(temp_df['货币(M1)-环比增长'])
        temp_df['流通中的现金(M0)-数量(亿元)'] = pd.to_numeric(temp_df['流通中的现金(M0)-数量(亿元)'])
        temp_df['流通中的现金(M0)-同比增长'] = pd.to_numeric(temp_df['流通中的现金(M0)-同比增长'])
        temp_df['流通中的现金(M0)-环比增长'] = pd.to_numeric(temp_df['流通中的现金(M0)-环比增长'])
        if start_date and end_date:
            start = start_date.split('-')[0] + '年' + start_date.split('-')[1] + '月份'
            end = end_date.split('-')[0] + '年' + end_date.split('-')[1] + '月份'
            temp_df = temp_df[(temp_df['月份'] >= start) & (temp_df['月份'] <= end)]
        return temp_df.to_markdown()

    @classmethod
    def china_consumer_goods_retail(cls, start_date: str = '', end_date: str = '') -> pd.DataFrame:
        """中国社会消费品零售总额月度统计数据。\
返回月份，当月总额（单位：亿元），同比增长（单位：%），\
环比增长（单位：%），当年累计（单位：亿元），累计同比增长（单位：%）
        """
        url = 'https://datacenter-web.eastmoney.com/api/data/v1/get'
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.138 Safari/537.36',  # noqa
        }
        params = {
            'columns': 'REPORT_DATE,TIME,RETAIL_TOTAL,RETAIL_TOTAL_SAME,RETAIL_TOTAL_SEQUENTIAL,RETAIL_TOTAL_ACCUMULATE,RETAIL_ACCUMULATE_SAME',  # noqa
            'pageNumber': '1',
            'pageSize': '1000',
            'sortColumns': 'REPORT_DATE',
            'sortTypes': '-1',
            'source': 'WEB',
            'client': 'WEB',
            'reportName': 'RPT_ECONOMY_TOTAL_RETAIL',
            'p': '1',
            'pageNo': '1',
            'pageNum': '1',
            '_': '1660718498421',
        }
        r = requests.get(url, params=params, headers=headers)
        data_json = r.json()
        temp_df = pd.DataFrame(data_json['result']['data'])
        temp_df.columns = [
            '-',
            '月份',
            '当月',
            '同比增长',
            '环比增长',
            '累计',
            '累计-同比增长',
        ]
        temp_df = temp_df[
            [
                '月份',
                '当月',
                '同比增长',
                '环比增长',
                '累计',
                '累计-同比增长',
            ]
        ]
        temp_df['当月'] = pd.to_numeric(temp_df['当月'], errors='coerce')
        temp_df['同比增长'] = pd.to_numeric(temp_df['同比增长'], errors='coerce')
        temp_df['环比增长'] = pd.to_numeric(temp_df['环比增长'], errors='coerce')
        temp_df['累计'] = pd.to_numeric(temp_df['累计'], errors='coerce')
        temp_df['累计-同比增长'] = pd.to_numeric(temp_df['累计-同比增长'], errors='coerce')
        if start_date and end_date:
            start = start_date.split('-')[0] + '年' + start_date.split('-')[1] + '月份'
            end = end_date.split('-')[0] + '年' + end_date.split('-')[1] + '月份'
            temp_df = temp_df[(temp_df['月份'] >= start) & (temp_df['月份'] <= end)]

        return temp_df.to_markdown()

    @classmethod
    def bond_zh_us_rate(cls, start_date: str = "", end_date: str = "") -> str:
        """
        本接口返回指定时间段[start_date,end_date]内交易日的中美两国的 2 年、5 年、10 年、30 年、10 年-2 年收益率数据。
        start_date表示起始日期，end_date表示结束日期，日期格式例如 2024-04-07
        """
        url = "https://datacenter.eastmoney.com/api/data/get"
        params = {
            "type": "RPTA_WEB_TREASURYYIELD",
            "sty": "ALL",
            "st": "SOLAR_DATE",
            "sr": "-1",
            "token": "894050c76af8597a853f5b408b759f5d",
            "p": "1",
            "ps": "500",
            "pageNo": "1",
            "pageNum": "1",
            "_": "1615791534490",
        }
        r = requests.get(url, params=params)
        data_json = r.json()
        total_page = data_json["result"]["pages"]
        big_df = pd.DataFrame()
        for page in range(1, total_page + 1):
            params = {
                "type": "RPTA_WEB_TREASURYYIELD",
                "sty": "ALL",
                "st": "SOLAR_DATE",
                "sr": "-1",
                "token": "894050c76af8597a853f5b408b759f5d",
                "p": page,
                "ps": "500",
                "pageNo": page,
                "pageNum": page,
                "_": "1615791534490",
            }
            r = requests.get(url, params=params)
            data_json = r.json()
            # 时间过滤
            if start_date and end_date:
                temp_data = []
                for item in data_json["result"]["data"]:
                    if start_date <= item["SOLAR_DATE"].split(" ")[0] <= end_date:
                        temp_data.append(item)
                    elif start_date > item["SOLAR_DATE"].split(" ")[0]:
                        break
                    else:
                        continue
            else:
                temp_data = data_json["result"]["data"]
            temp_df = pd.DataFrame(temp_data)
            for col in temp_df.columns:
                if temp_df[col].isnull().all():  # 检查列是否包含 None 或 NaN
                    temp_df[col] = pd.to_numeric(temp_df[col], errors='coerce')
            if big_df.empty:
                big_df = temp_df
            else:
                big_df = pd.concat(objs=[big_df, temp_df], ignore_index=True)

        big_df.rename(
            columns={
                "SOLAR_DATE": "日期",
                "EMM00166462": "中国国债收益率5年",
                "EMM00166466": "中国国债收益率10年",
                "EMM00166469": "中国国债收益率30年",
                "EMM00588704": "中国国债收益率2年",
                "EMM01276014": "中国国债收益率10年-2年",
                "EMG00001306": "美国国债收益率2年",
                "EMG00001308": "美国国债收益率5年",
                "EMG00001310": "美国国债收益率10年",
                "EMG00001312": "美国国债收益率30年",
                "EMG01339436": "美国国债收益率10年-2年",
                "EMM00000024": "中国GDP年增率",
                "EMG00159635": "美国GDP年增率",
            },
            inplace=True,
        )
        big_df = big_df[
            [
                "日期",
                "中国国债收益率2年",
                "中国国债收益率5年",
                "中国国债收益率10年",
                "中国国债收益率30年",
                "中国国债收益率10年-2年",
                "中国GDP年增率",
                "美国国债收益率2年",
                "美国国债收益率5年",
                "美国国债收益率10年",
                "美国国债收益率30年",
                "美国国债收益率10年-2年",
                "美国GDP年增率",
            ]
        ]
        big_df = big_df.drop(["中国GDP年增率", "美国GDP年增率"], axis=1)
        big_df["日期"] = pd.to_datetime(big_df["日期"], errors="coerce")
        big_df["中国国债收益率2年"] = pd.to_numeric(big_df["中国国债收益率2年"], errors="coerce")
        big_df["中国国债收益率5年"] = pd.to_numeric(big_df["中国国债收益率5年"], errors="coerce")
        big_df["中国国债收益率10年"] = pd.to_numeric(big_df["中国国债收益率10年"], errors="coerce")
        big_df["中国国债收益率30年"] = pd.to_numeric(big_df["中国国债收益率30年"], errors="coerce")
        big_df["中国国债收益率10年-2年"] = pd.to_numeric(big_df["中国国债收益率10年-2年"], errors="coerce")
        # big_df["中国GDP年增率"] = pd.to_numeric(big_df["中国GDP年增率"], errors="coerce")
        big_df["美国国债收益率2年"] = pd.to_numeric(big_df["美国国债收益率2年"], errors="coerce")
        big_df["美国国债收益率5年"] = pd.to_numeric(big_df["美国国债收益率5年"], errors="coerce")
        big_df["美国国债收益率10年"] = pd.to_numeric(big_df["美国国债收益率10年"], errors="coerce")
        big_df["美国国债收益率30年"] = pd.to_numeric(big_df["美国国债收益率30年"], errors="coerce")
        big_df["美国国债收益率10年-2年"] = pd.to_numeric(big_df["美国国债收益率10年-2年"], errors="coerce")
        # big_df["美国GDP年增率"] = pd.to_numeric(big_df["美国GDP年增率"], errors="coerce")
        big_df.sort_values("日期", inplace=True)
        big_df.set_index(["日期"], inplace=True)
        big_df = big_df[pd.to_datetime(start_date):]
        big_df.reset_index(inplace=True)
        big_df["日期"] = pd.to_datetime(big_df["日期"]).dt.date
        return big_df.to_markdown()

    @classmethod
    def get_api_tool(cls, name: str, **kwargs: Any) -> BaseTool:
        attr_name = name.split('_', 1)[-1]
        class_method = getattr(cls, attr_name)

        return MultArgsSchemaTool(name=name, description=class_method.__doc__, func=class_method, args_schema=QueryArg)


if __name__ == '__main__':
    tmp_start_date = '2024-01-01'
    tmp_end_date = '2024-01-03'
    # start_date = ''
    # end_date = ''
    # print(MacroData.china_ppi(start_date=start_date, end_date=end_date))
    # print(MacroData.china_shrzgm(start_date=start_date, end_date=end_date))
    # print(MacroData.china_consumer_goods_retail(start_date=start_date, end_date=end_date))
    # print(MacroData.china_cpi(start_date=start_date, end_date=end_date))
    # print(MacroData.china_pmi(start_date=start_date, end_date=end_date))
    # print(MacroData.china_money_supply(start_date=start_date, end_date=end_date))
    # print(MacroData.china_gdp_yearly(start_date=start_date, end_date=end_date))
    print(MacroData.bond_zh_us_rate(start_date=tmp_start_date, end_date=tmp_end_date))



from __future__ import annotations

import logging
# import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

# import numpy as np
from langchain.embeddings.base import Embeddings
from langchain.utils import get_from_dict_or_env
from langchain_core.pydantic_v1 import BaseModel, Extra, Field, root_validator
from requests.exceptions import HTTPError
from tenacity import (before_sleep_log, retry, retry_if_exception_type, stop_after_attempt,
                      wait_exponential)

logger = logging.getLogger(__name__)


def _create_retry_decorator(embeddings: WenxinEmbeddings) -> Callable[[Any], Any]:
    min_seconds = 4
    max_seconds = 10
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    return retry(
        reraise=True,
        stop=stop_after_attempt(embeddings.max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=(retry_if_exception_type(HTTPError)),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )


def embed_with_retry(embeddings: WenxinEmbeddings, **kwargs: Any) -> Any:
    """Use tenacity to retry the embedding call."""
    retry_decorator = _create_retry_decorator(embeddings)

    @retry_decorator
    def _embed_with_retry(**kwargs: Any) -> Any:
        return embeddings.embed(**kwargs)

    return _embed_with_retry(**kwargs)


class WenxinEmbeddings(BaseModel, Embeddings):
    """Wenxin embedding models.

    To use, the environment variable ``WENXIN_API_KEY`` and ``WENXIN_SECRET_KEY``
    set with your API key or pass it as a named parameter to the constructor.

    Example:
        .. code-block:: python
            from bisheng_langchain.embeddings import WenxinEmbeddings
            wenxin_embeddings = WenxinEmbeddings(
               wenxin_api_key="my-api-key",
               wenxin_secret_key='xxx')

    """

    client: Optional[Any]  #: :meta private:
    model: str = 'embedding-v1'

    deployment: Optional[str] = 'default'
    wenxin_api_key: Optional[str] = None
    wenxin_secret_key: Optional[str] = None

    embedding_ctx_length: Optional[int] = 6144
    """The maximum number of tokens to embed at once."""
    """Maximum number of texts to embed in each batch"""
    max_retries: Optional[int] = 6
    """Maximum number of retries to make when generating."""
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout in seconds for the OpenAPI request."""

    model_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        values['wenxin_api_key'] = get_from_dict_or_env(values, 'wenxin_api_key', 'WENXIN_API_KEY')
        values['wenxin_secret_key'] = get_from_dict_or_env(
            values,
            'wenxin_secret_key',
            'WENXIN_SECRET_KEY',
        )

        api_key = values['wenxin_api_key']
        sec_key = values['wenxin_secret_key']
        try:
            from .interface import WenxinEmbeddingClient
            values['client'] = WenxinEmbeddingClient(api_key=api_key, sec_key=sec_key)
        except AttributeError:
            raise ValueError('Try upgrading it with `pip install --upgrade requests`.')
        return values

    @property
    def _invocation_params(self) -> Dict:
        wenxin_args = {
            'model': self.model,
            'request_timeout': self.request_timeout,
            **self.model_kwargs,
        }

        return wenxin_args

    def embed(self, texts: List[str]) -> List[List[float]]:
        inp = {'input': texts, 'model': self.model}
        outp = self.client.create(**inp)
        if outp['status_code'] != 200:
            raise ValueError(f"Wenxin API returned an error: {outp['status_message']}")
        return [e['embedding'] for e in outp['data']]

    def embed_documents(self,
                        texts: List[str],
                        chunk_size: Optional[int] = 0) -> List[List[float]]:
        embeddings = embed_with_retry(self, texts=texts)
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """Call out to OpenAI's embedding endpoint for embedding query text.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """

        embeddings = embed_with_retry(self, texts=[text])
        return embeddings[0]



from typing import Any

from langchain_core.messages import FunctionMessage, ToolMessage


class LiberalFunctionMessage(FunctionMessage):
    content: Any


class LiberalToolMessage(ToolMessage):
    content: Any



from langchain_core.prompts import PromptTemplate
from langchain_core.prompts.chat import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)


pirate_preamble = """
## Task & Context
你是一个准确且可靠的知识库问答助手，能够借助上下文知识回答用户的问题。

## Style Guide
你需要遵循以下的规则来回答用户问题：
- 如果给定的上下文中没有找到相关的答案，就说你不知道，不要试图编造答案；
- 回答的语言跟用户的问题语言保持一致，比如用户的问题是中文，请用中文作答；用户的问题是英文，请用英文作答；
- 你需要根据上下文给出尽可能详细的回答；
"""

messages = [
    SystemMessagePromptTemplate.from_template(pirate_preamble),
    HumanMessagePromptTemplate.from_template("{question}"),
]
COHERE_CHAT_PROMPT = ChatPromptTemplate.from_messages(messages)


# system_content = """## Task & Context
# 你是一个准确且可靠的知识库问答助手，能够借助上下文知识回答用户的问题。

# ## Style Guide
# 你需要遵循以下的规则来回答用户问题：
# - 如果给定的上下文中没有找到相关的答案，就说你不知道，不要试图编造答案；
# - 回答的语言跟用户的问题语言保持一致，比如用户的问题是中文，请用中文作答；用户的问题是英文，请用英文作答；
# - 你需要根据上下文给出尽可能详细的回答；|<instruct>|Carefully perform the following instructions, in order, starting each with a new line.
# Firstly, Decide which of the retrieved documents are relevant to the user's last input by writing 'Relevant Documents:' followed by comma-separated list of document numbers. If none are relevant, you should instead write 'None'.
# Secondly, Decide which of the retrieved documents contain facts that should be cited in a good answer to the user's last input by writing 'Cited Documents:' followed a comma-separated list of document numbers. If you dont want to cite any of them, you should instead write 'None'.
# Thirdly, Write 'Answer:' followed by a response to the user's last input. Use the retrieved documents to help you. Do not insert any citations or grounding markup.
# Finally, Write 'Grounded answer:' followed by a response to the user's last input. Use the symbols <co: doc> and </co: doc> to indicate when a fact comes from a document in the search result, e.g <co: 0>my fact</co: 0> for a fact from document 0.|<documents>|{documents}"""

system_content = """## Task & Context
你是一个准确且可靠的知识库问答助手，能够借助上下文知识回答用户的问题。

## Style Guide
你需要遵循以下的规则来回答用户问题：
- 如果给定的上下文中没有找到相关的答案，就说你不知道，不要试图编造答案；
- 回答的语言跟用户的问题语言保持一致，比如用户的问题是中文，请用中文作答；用户的问题是英文，请用英文作答；
- 你需要根据上下文给出尽可能详细的回答；|<instruct>|Carefully perform the following instructions, in order, starting each with a new line.
Write 'Answer:' followed by a response to the user's last input. Use the retrieved documents to help you. Do not insert any citations or grounding markup.|<documents>|{documents}"""
messages = [
    SystemMessagePromptTemplate.from_template(system_content),
    HumanMessagePromptTemplate.from_template("{question}"),
]
COHERE_LOCAL_CHAT_PROMPT_RAG = ChatPromptTemplate.from_messages(messages)


# prompt_template = """<BOS_TOKEN><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|># Safety Preamble
# The instructions in this section override those in the task description and style guide sections. Don't answer questions that are harmful or immoral.

# # System Preamble
# ## Basic Rules
# You are a powerful conversational AI trained by Cohere to help people. You are augmented by a number of tools, and your job is to use and consume the output of these tools to best help the user. You will see a conversation history between yourself and a user, ending with an utterance from the user. You will then see a specific instruction instructing you what kind of response to generate. When you answer the user's requests, you cite your sources in your answers, according to those instructions.

# # User Preamble
# ## Task & Context
# 你是一个准确且可靠的知识库问答助手，能够借助上下文知识回答用户的问题。

# ## Style Guide
# 你需要遵循以下的规则来回答用户问题：
# - 如果给定的上下文中没有找到相关的答案，就说你不知道，不要试图编造答案；
# - 回答的语言跟用户的问题语言保持一致，比如用户的问题是中文，请用中文作答；用户的问题是英文，请用英文作答；
# - 你需要根据上下文给出尽可能详细的回答；<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|USER_TOKEN|>{question}<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|><results>
# {documents}
# </results><|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>Carefully perform the following instructions, in order, starting each with a new line.
# Write 'Answer:' followed by a response to the user's last input. Use the retrieved documents to help you. Do not insert any citations or grounding markup.<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>"""


prompt_template = """<BOS_TOKEN><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|># Safety Preamble
The instructions in this section override those in the task description and style guide sections. Don't answer questions that are harmful or immoral.

# System Preamble
## Basic Rules
You are a powerful conversational AI trained by Cohere to help people. You are augmented by a number of tools, and your job is to use and consume the output of these tools to best help the user. You will see a conversation history between yourself and a user, ending with an utterance from the user. You will then see a specific instruction instructing you what kind of response to generate.

# User Preamble
## Task & Context
你是一个准确且可靠的知识库问答助手，能够借助上下文知识回答用户的问题。

## Style Guide
Unless the user asks for a different style of answer, you should answer in full sentences, using proper grammar and spelling.<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|USER_TOKEN|>{question}<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|><results>
{documents}
</results><|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>Carefully perform the following instructions, in order, starting each with a new line.
- 回答的语言跟用户的问题语言保持一致，比如用户的问题是中文，请用中文作答；用户的问题是英文，请用英文作答；
- 你需要根据上下文给出尽可能详细的回答；
- Write 'Answer:' followed by a response to the user's last input. Use the retrieved documents to help you. <|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>"""
COHERE_RAW_PROMPT = PromptTemplate(
    template=prompt_template, input_variables=["documents", "question"]
)


# prompt_template = """<BOS_TOKEN><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>Use the following pieces of context to answer the question at the end. 
# If you don't know the answer, just say that you don't know, don't try to make up an answer.<results>
# {documents}
# </results><|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|USER_TOKEN|>{question}<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>"""

# # prompt_template = """<BOS_TOKEN><|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>Use the following pieces of context to answer the question at the end. 
# # If you don't know the answer, just say that you don't know, don't try to make up an answer.
# # ----------------
# # {documents}<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|USER_TOKEN|>{question}<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>"""
# COHERE_RAW_PROMPT = PromptTemplate(
#     template=prompt_template, input_variables=["documents", "question"]
# )


from typing import Any, Dict, List, Optional

import requests
from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import BaseModel, Extra, Field
DEFAULT_Multilingual_MODEL = "intfloat/multilingual-e5-large"


class HuggingFaceMultilingualEmbeddings(BaseModel, Embeddings):
    """HuggingFace BGE sentence_transformers embedding models.

    To use, you should have the ``sentence_transformers`` python package installed.

    Example:
        .. code-block:: python

            from langchain_community.embeddings import HuggingFaceBgeEmbeddings

            model_name = "BAAI/bge-large-en"
            model_kwargs = {'device': 'cpu'}
            encode_kwargs = {'normalize_embeddings': True}
            hf = HuggingFaceBgeEmbeddings(
                
                model_name=model_name,
                model_kwargs=model_kwargs,
                encode_kwargs=encode_kwargs
            )
    """

    client: Any  #: :meta private:
    model_name: str = DEFAULT_Multilingual_MODEL
    """Model name to use."""
    cache_folder: Optional[str] = None
    """Path to store models.
    Can be also set by SENTENCE_TRANSFORMERS_HOME environment variable."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Keyword arguments to pass to the model."""
    encode_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Keyword arguments to pass when calling the `encode` method of the model."""
    # query_instruction: str = DEFAULT_QUERY_BGE_INSTRUCTION_EN
    """Instruction to use for embedding query."""

    def __init__(self, **kwargs: Any):
        """Initialize the sentence_transformer."""
        super().__init__(**kwargs)
        try:
            import sentence_transformers

        except ImportError as exc:
            raise ImportError(
                "Could not import sentence_transformers python package. "
                "Please install it with `pip install sentence_transformers`."
            ) from exc

        self.client = sentence_transformers.SentenceTransformer(
            self.model_name, cache_folder=self.cache_folder, **self.model_kwargs
        )
        # if "-zh" in self.model_name:
        #     self.query_instruction = DEFAULT_QUERY_BGE_INSTRUCTION_ZH

    # class Config:
    #     """Configuration for this pydantic object."""

    #     extra = Extra.forbid

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a HuggingFace transformer model.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        texts = [t.replace("\n", " ") for t in texts]
        textschange = []
        for t in texts:
            textschange.append("passage: " + t)
        # print(textschange)
        
        embeddings = self.client.encode(textschange, **self.encode_kwargs)
        return embeddings.tolist()

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a HuggingFace transformer model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        text = text.replace("\n", " ")
        text = "query: " + text
        # print(text)
        embedding = self.client.encode(text, **self.encode_kwargs
        )
        return embedding.tolist()
    
    







import json
import os
import re
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from datasets import Dataset
from loguru import logger
from bisheng_ragas import evaluate
from bisheng_ragas.metrics import AnswerCorrectness, AnswerCorrectnessBisheng, AnswerRecallBisheng


@dataclass
class RagScore:
    excel_path: str
    save_path: str
    question_column: str
    gt_column: str
    answer_column: str
    metrics: List[str]
    contexts_column: Optional[str] = None
    query_type_column: Optional[str] = None
    gt_split_column: Optional[str] = None
    batch_size: int = 5
    whether_gtsplit: bool = False

    def _validate_metrics(self):
        for metric in self.metrics:
            if not hasattr(self, f'ragas_{metric}'):
                raise Exception(f'"ragas_{metric}" 未实现!')

    def ragas_answer_correctness(self, dataset: Dataset) -> pd.DataFrame:
        # answer_correctness, 只考虑事实相似度
        weights = [1.0, 0.0]
        answer_correctness = AnswerCorrectness(weights=weights, batch_size=self.batch_size)
        result = evaluate(
            dataset=dataset,
            metrics=[
                answer_correctness,
            ],
        )
        self.score_map_keys = list(result.keys())
        df = result.to_pandas()
        return df

    def ragas_answer_correctness_bisheng(self, dataset: Dataset) -> pd.DataFrame:
        answer_correctness = AnswerCorrectnessBisheng(batch_size=self.batch_size)
        result = evaluate(
            dataset=dataset,
            metrics=[answer_correctness],
        )
        self.score_map_keys = list(result.keys())
        df = result.to_pandas()
        return df

    def ragas_answer_recall_bisheng(self, dataset: Dataset) -> pd.DataFrame:
        answer_recall =AnswerRecallBisheng(batch_size=self.batch_size, 
                                           whether_gtsplit=self.whether_gtsplit)
        result = evaluate(
            dataset=dataset,
            metrics=[answer_recall],
        )
        self.score_map_keys = list(result.keys())
        df = result.to_pandas()
        return df

    def _remove_source(self, pred: str) -> str:
        """去除【1†source】, only for openai assistant"""
        pattern = re.compile("【(\d+)†source】")
        match = re.findall(pattern, pred)
        for i in match:
            str_temp = f"【{i}†source】"
            pred = pred.replace(str_temp, '')
        return pred

    def score(self) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        df = pd.read_excel(self.excel_path)
        ori_row_nums = df.shape[0]

        # 删除含有na的行
        columns_to_check = [
            self.question_column,
            self.gt_column,
            self.answer_column,
            self.contexts_column,
            self.query_type_column,
        ]
        # 是否有要点拆分列
        if self.gt_split_column:
            columns_to_check.append(self.gt_split_column)

        df.dropna(subset=[col for col in columns_to_check if col], inplace=True)
        df = df.reset_index()
        print(f'删除含有na的行 {ori_row_nums - df.shape[0]} 个!')
        print(f'总计 {df.shape[0]} 个问题')

        questions = df[self.question_column].tolist()
        answers = df[self.answer_column].tolist()
        # answers = df[self.answer_column].apply(self._remove_source).tolist() # for openai assistant
        ground_truths = df[self.gt_column].apply(lambda x: [x]).tolist()
        # todo: contexts可能是保存在json中的，这段代码可能需要修改
        contexts = (
            [['']] * len(questions)
            if not self.contexts_column
            else df[self.contexts_column].apply(lambda x: [x]).tolist()
        )
        # To dict
        if self.gt_split_column:
            gtsplit = df[self.gt_split_column].tolist()
            data: Dict[str, List[Any]] = {
                "question": questions,
                "answer": answers,
                "contexts": contexts,
                "ground_truths": ground_truths,
                'gt_split_point': gtsplit
            }
        else:
            data: Dict[str, List[Any]] = {
                "question": questions,
                "answer": answers,
                "contexts": contexts,
                "ground_truths": ground_truths,
            }
        # Convert dict to dataset
        dataset = Dataset.from_dict(data)

        self._validate_metrics()

        save_group_df = dict()
        for metric_name in self.metrics:
            ragas_result = getattr(self, f'ragas_{metric_name}')(dataset)
            if metric_name =='answer_recall_bisheng':
                if self.gt_split_column:
                    df[self.gt_split_column] = ragas_result["gt_split_point"]
                else:
                    df["gt_split_point"] = ragas_result["gt_split_point"]
                df["analyse"] = ragas_result["analyse"]

            score_map = dict().fromkeys(self.score_map_keys, ragas_result)
            for metric, scores in score_map.items():
                df[metric] = df.index.map({idx: rows[metric] for idx, rows in scores.iterrows()})

            if self.query_type_column and self.query_type_column in df.columns:
                grouped_df = df.groupby(self.query_type_column)
                grouped_df = grouped_df.agg({self.question_column: 'count', **{metric: 'mean' for metric in score_map}})
                grouped_df.rename(columns={self.question_column: '问题个数'}, inplace=True)
                
                total_question = grouped_df['问题个数'].sum()
                grouped_df.loc['总计', '问题个数'] = total_question
                for metric in score_map:
                    grouped_df.loc['总计', metric] = df[metric].sum() / total_question
                save_group_df[f'{metric_name}_group'] = grouped_df
                
                print(grouped_df.to_markdown())

        # save
        output_path = Path(self.save_path) / f"{Path(self.excel_path).stem}_score.xlsx"
        with pd.ExcelWriter(output_path) as writer:
            df.to_excel(writer, sheet_name='Sheet1', index=False)
            if len(save_group_df):
                for metric, grouped_df in save_group_df.items():
                    grouped_df.to_excel(writer, sheet_name=metric, index=True)
        print(f'保存到 {output_path} 成功!')


if __name__ == '__main__':
    params = {
        'excel_path': '/home/gulixin/workspace/llm/bisheng/src/bisheng-langchain/experimental/rag/data/test.xlsx',
        'save_path': '/home/gulixin/workspace/llm/bisheng/src/bisheng-langchain/experimental/rag/data',
        'question_column': '问题',
        'gt_column': 'GT',
        'answer_column': 'rag_answer',
        'query_type_column': '问题类型',
        # 'metrics': ['answer_correctness_bisheng'],
        'metrics': ['answer_recall_bisheng'],
        'batch_size': 10,
        'whether_gtsplit': False,
    }
    rag_score = RagScore(**params)
    rag_score.score()



from langchain_core.prompts import PromptTemplate
from langchain_core.prompts.chat import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)

# 例如助手的身份、相关完成任务的具体方法和步骤、回答问题时的语气以及应该注意什么问题等
system_template = """你是一个prompt优化大师，你会得到一个助手的名字和简单描述，你需要根据这些信息，为助手生成一个合适的角色描述、详细的技能说明、相关约束信息，输出为markdown格式。你需要按照以下格式进行组织输出内容：
```markdown
## 角色
你是一个[助手的角色]，[助手的角色描述]。

## 技能
1. [技能 1 的描述]：
  - [技能 1 的具体内容]。
  - [技能 1 的具体内容]。
2. [技能 2 的描述]：
  - [技能 2 的具体内容]。
  - [技能 2 的具体内容]。

## 限制
- [限制 1 的描述]。
- [限制 2 的描述]。
```

以下是一些例子：
示例1：
输入：
助手名字: 金融分析助手
助手描述: 1. 分析上市公司最新的年报财报；2. 获取上市公司的最新新闻；

输出：
```markdown
## 角色
你是一个金融分析师，会利用最新的信息和数据来分析公司的财务状况、市场趋势和行业动态，以帮助客户做出明智的投资决策。

## 技能
1. 分析上市公司最新的年报财报：
  - 使用财务分析工具和技巧，对公司的财务报表进行详细的分析和解读。
  - 评估公司的财务健康状况，包括营收、利润、资产负债表、现金流量等方面。
  - 分析公司的财务指标，如利润率、偿债能力、周转率等，以评估其盈利能力和风险水平。
  - 比较公司的财务表现与同行业其他公司的平均水平，以评估其相对竞争力。
2. 获取上市公司的最新新闻：
  - 使用新闻来源和数据库，定期获取上市公司的最新新闻和公告。
  - 分析新闻对公司股价和投资者情绪的潜在影响。
  - 关注公司的重大事件，如合并收购、产品发布、管理层变动等，以及这些事件对公司未来发展的影响。
  - 结合财务分析和新闻分析，提供对公司的综合评估和投资建议。

## 限制
- 只讨论与金融分析相关的内容，拒绝回答与金融分析无关的话题。
- 所有的输出内容必须按照给定的格式进行组织，不能偏离框架要求。
- 分析部分不能超过 100 字。
```

示例2：
输入：
助手名字: 前端开发助手
助手描述: 你的角色是前端开发，能帮助我把图片制作成html页面，css使用tailwind.css，ui库使用antd

输出：
```markdown
# 角色
你是一个前端开发工程师，可以使用 HTML、CSS 和 JavaScript 等技术构建网站和应用程序。

## 技能
1. 将图片制作成 HTML 页面
  - 当用户需要将图片制作成 HTML 页面时，你可以根据用户提供的图片和要求，使用 HTML 和 CSS 等技术构建一个页面。
  - 在构建页面时，你可以使用 Tailwind CSS 来简化 CSS 样式的编写，并使用 Antd 库来提供丰富的 UI 组件。
  - 构建完成后，你可以将页面代码返回给用户，以便用户可以将其部署到服务器上或在本地查看。

2. 提供前端开发相关的建议和帮助
  - 当用户需要前端开发相关的建议和帮助时，你可以根据用户的问题，提供相关的建议和帮助。
  - 你可以提供关于 HTML、CSS、JavaScript 等前端技术的建议和帮助，也可以提供关于前端开发工具和流程的建议和帮助。

## 限制
- 只讨论与前端开发相关的内容，拒绝回答与前端开发无关的话题。
- 所输出的内容必须按照给定的格式进行组织，不能偏离框架要求。
```
"""

human_template = """
输入：
助手名字: {assistant_name}
助手描述: {assistant_description}

输出：
"""


messages = [
    SystemMessagePromptTemplate.from_template(system_template),
    HumanMessagePromptTemplate.from_template(human_template),
]
ASSISTANT_PROMPT_OPT = ChatPromptTemplate.from_messages(messages)